import typing as t 
import jax.numpy as jnp
import jax.random as rand

def get_ray_bundle(height: int, width: int, focal_length: float, cam2world: jnp.array) -> t.Tuple[jnp.array, jnp.array]:
    """Generate ray bundle expressed in world coordinates

    Args:
        height (int): _description_
        width (int): _description_
        focal_length (float): _description_
        cam2world (jnp.array): _description_

    Returns:
        t.Tuple[jnp.array, jnp.array]: _description_
    """
    xx, yy = jnp.meshgrid(jnp.arange(width), jnp.arange(height), indexing="xy")
    directions_c = jnp.stack([(xx - .5 * width) / focal_length, 
                              (yy - .5 * height) / focal_length,
                              -jnp.ones_like(xx)], axis=-1) # HxWx3
    directions_w =  jnp.dot(directions_c, cam2world[:3, :3])  #HxWx3
    origins_w =  jnp.broadcast_to(cam2world[:3, -1], directions_c.shape) # HxWx3
    return origins_w, directions_w

def sample_query_points(origins: jnp.array,
                        directions: jnp.array,
                        near_thres: float,
                        far_thres: float, 
                        num_samples: int,
                        rand_key: t.Optional[rand.KeyArray] = None) -> t.Tuple[jnp.array, jnp.array]:
    """_summary_

    Args:
        origins (jnp.array): _description_
        directions (jnp.array): _description_
        near_thres (float): _description_
        far_thres (float): _description_
        num_samples (int): _description_
        rand_key (t.Optional[rand.KeyArray], optional): _description_. Defaults to None.

    Returns:
        t.Tuple[jnp.array, jnp.array]: _description_
    """
    depth_values = jnp.linspace(near_thres, far_thres, num_samples)  # (num_samples, )
    if rand_key is not None:
        # (H, W, num_samples) = (1, 1, num_samples) + (H, W, num_samples)
        depth_values = depth_values[jnp.newaxis, jnp.newaxis, :] + rand.uniform(rand_key, 
                                                                    shape=(*origins.shape[:-1], num_samples)) / num_samples
    else:
        depth_values = jnp.expand_dims(depth_values, axis=(0, 1))
    # (H, W, num_samples, 3) = (H, W, 1, 3) + (H, W, 1, 3) * (1, 1, num_samples, 1) or (H, W, num_samples, 1)
    query_points = origins[..., jnp.newaxis, :] + directions[..., jnp.newaxis, :] * depth_values[..., jnp.newaxis]
    return query_points, depth_values

def positional_encoding(query_points: jnp.array, num_encodings: int) -> jnp.array:
    """ Create sinusioidal positional encodings from query_points

    Args:
        query_points (jnp.array): query_points flattened => (H*W*num_samples, 3)
        num_encodings (int): number of encodings/frequency bands

    Returns:
        jnp.array: encoded feature vec (H*W*num_samples, 3*num_encodings)
    """
    encodings = []
    freq_bands = 2**jnp.linspace(0, num_encodings, num_encodings)
    for freq in freq_bands:
        for fn in [jnp.sin, jnp.cos]:
            encodings.append(fn(freq * query_points))
    encodings = jnp.concatenate(encodings, axis=-1)
    return encodings



import typing as t
from flax import linen as nn

class MLP(nn.Module):
    features: t.Sequence[int]
    output_dim: int

    @nn.compact
    def __call__(self, x):
        for features in self.features:
            x = nn.Dense(features)(x)
            x = nn.relu(x)
        x = nn.Dense(self.output_dim)(x)
        x = nn.relu(x)
        return x

class MiniResNet(nn.Module):
    features: t.Sequence[int]
    output_dim: int

    @nn.compact
    def __call__(self, x):
        for features in self.features:
            x_ = nn.Dense(features)(x)
            x = nn.relu(x_) + x 
        x = nn.Dense(self.output_dim)(x)
        return x 

class CNN(nn.Module):
    features: t.Sequence[int]
    kernel_sizes: t.Sequence[t.Sequence[int]]
    output_dim: int 

    @nn.compact
    def __call__(self, x):
        for features, kernel_size in zip(self.features, self.kernel_sizes):
            x = nn.Conv(features=features, kernel_size=kernel_size)(x)
            x = nn.relu(x)
        x = nn.Dense(self.output_dim)(x)
        return x



from pathlib import Path
import typing as t
import numpy as np

def load_image_data(image_data_path: Path) -> t.Dict[str, np.ndarray]:
    data = np.load(image_data_path)
    images = data["images"]
    cam2world = data["poses"]
    focal_length = data["focal"]
    return images, cam2world,  focal_length


import typing as t
import jax.numpy as jnp

def exc_cumprod(factors: jnp.ndarray) -> jnp.ndarray:
    cumprod = jnp.cumprod(factors, axis=-1)
    cumprod = jnp.roll(cumprod, shift=1, axis=-1)
    cumprod = cumprod.at[0].set(1.0)
    return cumprod

def render_volume_density(radience_field: jnp.ndarray,
                          depth_values: jnp.ndarray) -> t.Tuple[jnp.ndarray]:
    """ Render rgb & depth from radiance field """
    sigma = radience_field[..., 3]
    rgb = radience_field[..., :3]

    one_e_10 = jnp.array([1e10])
    dist = jnp.concatenate((depth_values[..., 1:] - depth_values[..., :-1], 
                            jnp.broadcast_to(one_e_10, depth_values.shape[:2] + (1,))), axis=-1)
    alpha = 1. - jnp.exp(- sigma * dist)
    weights = alpha  * exc_cumprod(1. - alpha + 1e-10)

    color = jnp.sum(weights[..., jnp.newaxis] * rgb, axis=-2)
    depth = jnp.sum(weights * depth_values, axis=-1)
    return color, depth


import typing as t
import argparse
from pathlib import Path
from functools import partial

import jax.random as rd
import jax.numpy as jnp
import numpy as np
import flax 
import jax
import optax
import matplotlib.pyplot as plt
from flax.training import train_state
from tqdm import tqdm

from nerf.ray_helpers import get_ray_bundle, sample_query_points, positional_encoding
from nerf.model import MLP
from nerf.rendering import render_volume_density
from nerf.dataloading import load_image_data

class TrainState(train_state.TrainState):
    rngs: rd.KeyArray = None

def create_train_state(module: flax.linen.Module,
                       lr: float, 
                       pseudo_input: t.Optional[jnp.ndarray],
                       near_thres: float,
                       far_thres: float,
                       num_samples: int,
                       num_encodings: int,
                       rng: rd.KeyArray) -> TrainState:
    rngs = rd.split(rng, num=3)
    params = module.init(rngs[0], pseudo_input)["params"]
    tx = optax.adam(learning_rate=lr)
    opt_state = tx.init(params)
    return TrainState(apply_fn=partial(nerf_predict,
                                       near_thres=near_thres, 
                                       far_thres=far_thres, 
                                       num_samples=num_samples,
                                       num_encodings=num_encodings),
                      params=params,
                      tx=tx,
                      rngs=rngs[1:],
                      opt_state=opt_state,
                      step=0
                      )

def nerf_predict(params: jnp.ndarray,
                 image: jnp.ndarray, 
                 cam2world: jnp.array,
                 focal_length: float,
                 model: flax.linen.Module,
                 near_thres: float,
                 far_thres: float,
                 num_samples: int,
                 num_encodings: int,
                 batch_size:int,
                 rng: rd.KeyArray) -> jnp.ndarray:
    """Run one iteration of nerf"""
    new_key, _ = rd.split(key=rng)
    height, width, _ = image.shape
    ray_origins, ray_dirs = get_ray_bundle(height=height, 
                                            width=width, 
                                            focal_length=focal_length, 
                                            cam2world=cam2world)
                            
    query_points, depth_values = sample_query_points(origins=ray_origins, 
                                       directions=ray_dirs,
                                       near_thres=near_thres, 
                                       far_thres=far_thres, 
                                       num_samples=num_samples, 
                                       rand_key=new_key)
    enc_points = positional_encoding(query_points.reshape(-1, 3), num_encodings=num_encodings)
    
    predictions = []
    for i in tqdm(range(0, enc_points.shape[0], batch_size), desc="RENDERING"):
        batch = enc_points[i:i+batch_size, :]
        pred = model.apply({"params": params}, batch)
        predictions.append(pred)
    predictions = jnp.concatenate(predictions, axis=0)
    predictions = predictions.reshape(query_points.shape[:-1] + (4,))

    rgb_predicted, _ = render_volume_density(radience_field=predictions,
                                          depth_values=depth_values)

    return rgb_predicted



def train_step(state: TrainState,
               image: jnp.ndarray, 
               cam2world: jnp.array,
               focal_length,
               model: flax.linen.Module,
               batch_size: int):
    "Run one training step"
    def nerf_loss(params) -> float:
        predicted_img = state.apply_fn(params=params, 
                                       image=image, 
                                       cam2world=cam2world,
                                       focal_length=focal_length,
                                       model=model, 
                                       batch_size=batch_size, 
                                       rng=state.rngs[0])

        return jnp.sum((image - predicted_img) ** 2) / 2
    grad_fn = jax.grad(nerf_loss)
    grads = grad_fn(state.params)
    state = state.apply_gradients(grads=grads)
    state.replace(rngs=rd.split(state.rngs[1]))
    return state


def main():
    parser = argparse.ArgumentParser()
    #FOLDERS
    data_group = parser.add_argument_group("data")
    data_group.add_argument("--data_path", type=str, default="./data/tiny_nerf_data.npz")
    data_group.add_argument("--res_dir", type=str, default="./results")
    data_group.add_argument("--fig_dir", type=str, default="./results/figs")

    # NERF
    nerf_group = parser.add_argument_group("nerf")
    nerf_group.add_argument("--near_far_bounds", nargs=2, type=float, default=[1., 9.])
    nerf_group.add_argument("--num_encodings", default=2)
    nerf_group.add_argument("--samples_per_ray", type=int, default=2)
    
    # TRAIN
    train_group = parser.add_argument_group("train")
    train_group.add_argument("--batch_size", type=int, default=2)
    train_group.add_argument("--lr", type=float, default=1e-3)
    train_group.add_argument("--n_iters", type=int, default=3000)

    # OTHER 
    other_group = parser.add_argument_group("other")
    other_group.add_argument("--seed", type=int, default=42)

    args = parser.parse_args()

    # FOLDER OPS
    data_path = Path(args.data_path)
    assert data_path.exists(), f"No data found under {data_path.resolve()}"

    res_dir = Path(args.res_dir)
    fig_dir = Path(args.fig_dir)
    to_create = [res_dir, fig_dir]
    for folder in to_create:
        folder.mkdir(exist_ok=True, parents=True)
    
    # SET SEED
    rng_key = rd.PRNGKey(args.seed)
    
    # LOAD IMAGES
    images, cam2world, focal = load_image_data(data_path)
    image = images[0]
    
    # INIT MODEL
    model = MLP([128, 128], 4)
    h, w = image.shape[:2]
    pseudo_input = jnp.ones([h, w, 3*2*args.num_encodings])

    # Optimizer State and Model State
    rng_keys = rd.split(rng_key, num=3)
    state = create_train_state(module=model,
                              lr=args.lr, 
                              pseudo_input=pseudo_input,
                              rng=rng_keys[0],
                              near_thres=args.near_far_bounds[0],
                              far_thres=args.near_far_bounds[1],
                              num_samples=args.samples_per_ray,
                              num_encodings=args.num_encodings)

    for i in tqdm(range(args.n_iters), desc="TRAINING"):
        target_img_idx = rd.randint(rng_keys[1], (1,) , 0, images.shape[0]).item()
        target_img = images[target_img_idx]
        cam2world = cam2world[target_img_idx]

        state = train_step(state, 
                           target_img,
                           cam2world,
                           focal,
                           model,
                           batch_size=args.batch_size
                          )
        if i % 100 == 0 or i == args.n_iters - 1:
            predicted_img = nerf_predict(state.params,
                                         image,
                                         cam2world,
                                         focal,
                                         model,
                                         batch_size=args.batch_size,
                                         rng=rng_keys[2])
            plt.imshow(predicted_img)
            plt.savefig(fig_dir/ f"{str(i).zfill(10)}.png")
    

if __name__ == "__main__":
    main()


import sys

import jax.numpy as jnp
import jax.random as rnd
from nerf import get_ray_bundle, sample_query_points, positional_encoding

def test_ray_bundle(camera_intrinsics: jnp.array) -> None:
    height, width, focal_length = 100, 100, 2.
    origins, directions = get_ray_bundle(height, width, focal_length, camera_intrinsics)
    assert origins.shape == (height, width, 3)
    assert directions.shape == (height, width, 3)

def test_query_points(camera_intrinsics: jnp.array) -> None:
    height, width, focal_length = 100, 100, 2.
    origins, directions = get_ray_bundle(height, width, focal_length, camera_intrinsics)
    near_thres, far_thres, num_samples  = 0.1, 1.0, 100
    rng_key = rnd.PRNGKey(42)
    query_points, depth_values = sample_query_points(origins, 
                                                    directions,
                                                    near_thres, 
                                                    far_thres, 
                                                    num_samples,
                                                    rng_key)
    assert query_points.shape == (height, width, 3, num_samples)
    assert depth_values.shape == (height, width, 1, num_samples) 

def test_encoding_fn(camera_intrinsics: jnp.array) -> None:
    height, width, focal_length = 100, 100, 2.
    origins, directions = get_ray_bundle(height, width, focal_length, camera_intrinsics)
    near_thres, far_thres, num_samples  = 0.1, 1.0, 100
    rng_key = rnd.PRNGKey(42)
    query_points, depth_values = sample_query_points(origins, 
                                                    directions,
                                                    near_thres, 
                                                    far_thres, 
                                                    num_samples,
                                                    rng_key)
    query_points = jnp.reshape(query_points, (-1, 3))

    num_encodings = 4
    encodings = positional_encoding(query_points, num_encodings=num_encodings)
    assert encodings.shape == (height * width * num_samples, 3, num_encodings * 2)

